import logging
from typing import Optional

from geopy.distance import geodesic
from loguru import logger
from oslo_config import cfg

from aprsd import utils
from aprsd.packets.core import AckPacket, GPSPacket, RejectPacket


LOG = logging.getLogger()
LOGU = logger
CONF = cfg.CONF

FROM_COLOR = "fg #C70039"
TO_COLOR = "fg #D033FF"
TX_COLOR = "red"
RX_COLOR = "green"
PACKET_COLOR = "cyan"
DISTANCE_COLOR = "fg #FF5733"
DEGREES_COLOR = "fg #FFA900"


def log_multiline(packet, tx: Optional[bool] = False, header: Optional[bool] = True) -> None:
    """LOG a packet to the logfile."""
    if not CONF.enable_packet_logging:
        return
    if CONF.log_packet_format == "compact":
        return

    # asdict(packet)
    logit = ["\n"]
    name = packet.__class__.__name__

    if isinstance(packet, AckPacket):
        pkt_max_send_count = CONF.default_ack_send_count
    else:
        pkt_max_send_count = CONF.default_packet_send_count

    if header:
        if tx:
            header_str = f"<{TX_COLOR}>TX</{TX_COLOR}>"
            logit.append(
                f"{header_str}________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>  "
                f"TX:{packet.send_count + 1} of {pkt_max_send_count}",
            )
        else:
            header_str = f"<{RX_COLOR}>RX</{RX_COLOR}>"
            logit.append(
                f"{header_str}________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>)",
            )

    else:
        header_str = ""
        logit.append(f"__________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>)")
    # log_list.append(f"  Packet  : {packet.__class__.__name__}")
    if packet.msgNo:
        logit.append(f"  Msg #   : {packet.msgNo}")
    if packet.from_call:
        logit.append(f"  From    : <{FROM_COLOR}>{packet.from_call}</{FROM_COLOR}>")
    if packet.to_call:
        logit.append(f"  To      : <{TO_COLOR}>{packet.to_call}</{TO_COLOR}>")
    if hasattr(packet, "path") and packet.path:
        logit.append(f"  Path    : {'=>'.join(packet.path)}")
    if hasattr(packet, "via") and packet.via:
        logit.append(f"  VIA     : {packet.via}")

    if not isinstance(packet, AckPacket) and not isinstance(packet, RejectPacket):
        msg = packet.human_info

        if msg:
            msg = msg.replace("<", "\\<")
            logit.append(f"  Info    : <light-yellow><b>{msg}</b></light-yellow>")

    if hasattr(packet, "comment") and packet.comment:
        logit.append(f"  Comment : {packet.comment}")

    raw = packet.raw.replace("<", "\\<")
    logit.append(f"  Raw     : <fg #828282>{raw}</fg #828282>")
    logit.append(f"{header_str}________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>)")

    LOGU.opt(colors=True).info("\n".join(logit))
    LOG.debug(repr(packet))


def log(packet, tx: Optional[bool] = False, header: Optional[bool] = True) -> None:
    if not CONF.enable_packet_logging:
        return
    if CONF.log_packet_format == "multiline":
        log_multiline(packet, tx, header)
        return

    logit = []
    name = packet.__class__.__name__
    if isinstance(packet, AckPacket):
        pkt_max_send_count = CONF.default_ack_send_count
    else:
        pkt_max_send_count = CONF.default_packet_send_count

    if header:
        if tx:
            via_color = "red"
            arrow = f"<{via_color}>\u2192</{via_color}>"
            logit.append(
                f"<red>TX\u2191</red> "
                f"<cyan>{name}</cyan>"
                f":{packet.msgNo}"
                f" ({packet.send_count + 1} of {pkt_max_send_count})",
            )
        else:
            via_color = "fg #1AA730"
            arrow = f"<{via_color}>\u2192</{via_color}>"
            f"<{via_color}><-</{via_color}>"
            logit.append(
                f"<fg #1AA730>RX\u2193</fg #1AA730> "
                f"<cyan>{name}</cyan>"
                f":{packet.msgNo}",
            )
    else:
        via_color = "green"
        arrow = f"<{via_color}>-></{via_color}>"
        logit.append(
            f"<cyan>{name}</cyan>"
            f":{packet.msgNo}",
        )

    tmp = None
    if packet.path:
        tmp = f"{arrow}".join(packet.path) + f"{arrow} "

    logit.append(
        f"<{FROM_COLOR}>{packet.from_call}</{FROM_COLOR}> {arrow}"
        f"{tmp if tmp else ' '}"
        f"<{TO_COLOR}>{packet.to_call}</{TO_COLOR}>",
    )

    if not isinstance(packet, AckPacket) and not isinstance(packet, RejectPacket):
        logit.append(":")
        msg = packet.human_info

        if msg:
            msg = msg.replace("<", "\\<")
            logit.append(f"<light-yellow><b>{msg}</b></light-yellow>")

    # is there distance information?
    if isinstance(packet, GPSPacket) and CONF.latitude and CONF.longitude:
        my_coords = (CONF.latitude, CONF.longitude)
        packet_coords = (packet.latitude, packet.longitude)
        try:
            bearing = utils.calculate_initial_compass_bearing(my_coords, packet_coords)
        except Exception as e:
            LOG.error(f"Failed to calculate bearing: {e}")
            bearing = 0
        logit.append(
            f" : <{DEGREES_COLOR}>{utils.degrees_to_cardinal(bearing, full_string=True)}</{DEGREES_COLOR}>"
            f"<{DISTANCE_COLOR}>@{geodesic(my_coords, packet_coords).miles:.2f}miles</{DISTANCE_COLOR}>",
        )

    LOGU.opt(colors=True).info(" ".join(logit))
    log_multiline(packet, tx, header)



import datetime
import logging
import time

from aprslib.exceptions import LoginError
from oslo_config import cfg

from aprsd import client, exception
from aprsd.client import base
from aprsd.client.drivers import aprsis
from aprsd.packets import core


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class APRSISClient(base.APRSClient):

    _client = None

    def __init__(self):
        max_timeout = {"hours": 0.0, "minutes": 2, "seconds": 0}
        self.max_delta = datetime.timedelta(**max_timeout)

    def stats(self) -> dict:
        stats = {}
        if self.is_configured():
            stats = {
                "server_string": self._client.server_string,
                "sever_keepalive": self._client.aprsd_keepalive,
                "filter": self.filter,
            }

        return stats

    @staticmethod
    def is_enabled():
        # Defaults to True if the enabled flag is non existent
        try:
            return CONF.aprs_network.enabled
        except KeyError:
            return False

    @staticmethod
    def is_configured():
        if APRSISClient.is_enabled():
            # Ensure that the config vars are correctly set
            if not CONF.aprs_network.login:
                LOG.error("Config aprs_network.login not set.")
                raise exception.MissingConfigOptionException(
                    "aprs_network.login is not set.",
                )
            if not CONF.aprs_network.password:
                LOG.error("Config aprs_network.password not set.")
                raise exception.MissingConfigOptionException(
                    "aprs_network.password is not set.",
                )
            if not CONF.aprs_network.host:
                LOG.error("Config aprs_network.host not set.")
                raise exception.MissingConfigOptionException(
                    "aprs_network.host is not set.",
                )

            return True
        return True

    def _is_stale_connection(self):
        delta = datetime.datetime.now() - self._client.aprsd_keepalive
        if delta > self.max_delta:
            LOG.error(f"Connection is stale, last heard {delta} ago.")
            return True

    def is_alive(self):
        if self._client:
            return self._client.is_alive() and not self._is_stale_connection()
        else:
            LOG.warning(f"APRS_CLIENT {self._client} alive? NO!!!")
            return False

    def close(self):
        if self._client:
            self._client.stop()
            self._client.close()

    @staticmethod
    def transport():
        return client.TRANSPORT_APRSIS

    def decode_packet(self, *args, **kwargs):
        """APRS lib already decodes this."""
        return core.factory(args[0])

    def setup_connection(self):
        user = CONF.aprs_network.login
        password = CONF.aprs_network.password
        host = CONF.aprs_network.host
        port = CONF.aprs_network.port
        self.connected = False
        backoff = 1
        aprs_client = None
        while not self.connected:
            try:
                LOG.info(f"Creating aprslib client({host}:{port}) and logging in {user}.")
                aprs_client = aprsis.Aprsdis(user, passwd=password, host=host, port=port)
                # Force the log to be the same
                aprs_client.logger = LOG
                aprs_client.connect()
                self.connected = True
                backoff = 1
            except LoginError as e:
                LOG.error(f"Failed to login to APRS-IS Server '{e}'")
                self.connected = False
                time.sleep(backoff)
            except Exception as e:
                LOG.error(f"Unable to connect to APRS-IS server. '{e}' ")
                self.connected = False
                time.sleep(backoff)
                # Don't allow the backoff to go to inifinity.
                if backoff > 5:
                    backoff = 5
                else:
                    backoff += 1
                continue
        self._client = aprs_client
        return aprs_client

    def consumer(self, callback, blocking=False, immortal=False, raw=False):
        self._client.consumer(
            callback, blocking=blocking,
            immortal=immortal, raw=raw,
        )



from oslo_config import cfg


email_group = cfg.OptGroup(
    name="email_plugin",
    title="Options for the APRSD Email plugin",
)

email_opts = [
    cfg.StrOpt(
        "callsign",
        help="(Required) Callsign to validate for doing email commands."
             "Only this callsign can check email. This is also where the "
             "email notifications for new emails will be sent.",
    ),
    cfg.BoolOpt(
        "enabled",
        default=False,
        help="Enable the Email plugin?",
    ),
    cfg.BoolOpt(
        "debug",
        default=False,
        help="Enable the Email plugin Debugging?",
    ),
]

email_imap_opts = [
    cfg.StrOpt(
        "imap_login",
        help="Login username/email for IMAP server",
    ),
    cfg.StrOpt(
        "imap_password",
        secret=True,
        help="Login password for IMAP server",
    ),
    cfg.HostnameOpt(
        "imap_host",
        help="Hostname/IP of the IMAP server",
    ),
    cfg.PortOpt(
        "imap_port",
        default=993,
        help="Port to use for IMAP server",
    ),
    cfg.BoolOpt(
        "imap_use_ssl",
        default=True,
        help="Use SSL for connection to IMAP Server",
    ),
]

email_smtp_opts = [
    cfg.StrOpt(
        "smtp_login",
        help="Login username/email for SMTP server",
    ),
    cfg.StrOpt(
        "smtp_password",
        secret=True,
        help="Login password for SMTP server",
    ),
    cfg.HostnameOpt(
        "smtp_host",
        help="Hostname/IP of the SMTP server",
    ),
    cfg.PortOpt(
        "smtp_port",
        default=465,
        help="Port to use for SMTP server",
    ),
    cfg.BoolOpt(
        "smtp_use_ssl",
        default=True,
        help="Use SSL for connection to SMTP Server",
    ),
]

email_shortcuts_opts = [
    cfg.ListOpt(
        "email_shortcuts",
        help="List of email shortcuts for checking/sending email "
             "For Exmaple: wb=walt@walt.com,cl=cl@cl.com\n"
             "Means use 'wb' to send an email to walt@walt.com",
    ),
]

ALL_OPTS = (
    email_opts
    + email_imap_opts
    + email_smtp_opts
    + email_shortcuts_opts
)


def register_opts(config):
    config.register_group(email_group)
    config.register_opts(ALL_OPTS, group=email_group)


def list_opts():
    return {
        email_group.name: ALL_OPTS,
    }



from unittest import mock

from oslo_config import cfg

from aprsd import client, packets
from aprsd import conf  # noqa: F401
from aprsd.plugins import notify as notify_plugin

from .. import fake, test_plugin


CONF = cfg.CONF
DEFAULT_WATCHLIST_CALLSIGNS = fake.FAKE_FROM_CALLSIGN


class TestWatchListPlugin(test_plugin.TestPlugin):
    def setUp(self):
        self.fromcall = fake.FAKE_FROM_CALLSIGN
        self.ack = 1

    def config_and_init(
        self,
        watchlist_enabled=True,
        watchlist_alert_callsign=None,
        watchlist_alert_time_seconds=None,
        watchlist_packet_keep_count=None,
        watchlist_callsigns=DEFAULT_WATCHLIST_CALLSIGNS,
    ):
        CONF.callsign = self.fromcall
        CONF.aprs_network.login = self.fromcall
        CONF.aprs_fi.apiKey = "something"

        # Set the watchlist specific config options
        CONF.watch_list.enabled = watchlist_enabled

        if not watchlist_alert_callsign:
            watchlist_alert_callsign = fake.FAKE_TO_CALLSIGN
        CONF.watch_list.alert_callsign = watchlist_alert_callsign

        if not watchlist_alert_time_seconds:
            watchlist_alert_time_seconds = CONF.watch_list.alert_time_seconds
        CONF.watch_list.alert_time_seconds = watchlist_alert_time_seconds

        if not watchlist_packet_keep_count:
            watchlist_packet_keep_count = CONF.watch_list.packet_keep_count
            CONF.watch_list.packet_keep_count = watchlist_packet_keep_count

        CONF.watch_list.callsigns = watchlist_callsigns


class TestAPRSDWatchListPluginBase(TestWatchListPlugin):

    def test_watchlist_not_enabled(self):
        self.config_and_init(watchlist_enabled=False)
        plugin = fake.FakeWatchListPlugin()

        packet = fake.fake_packet(
            message="version",
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.client.factory.ClientFactory", autospec=True)
    def test_watchlist_not_in_watchlist(self, mock_factory):
        client.client_factory = mock_factory
        self.config_and_init()
        plugin = fake.FakeWatchListPlugin()

        packet = fake.fake_packet(
            fromcall="FAKE",
            message="version",
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)


class TestNotifySeenPlugin(TestWatchListPlugin):

    def test_disabled(self):
        self.config_and_init(watchlist_enabled=False)
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            message="version",
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.client.factory.ClientFactory", autospec=True)
    def test_callsign_not_in_watchlist(self, mock_factory):
        client.client_factory = mock_factory
        self.config_and_init(watchlist_enabled=False)
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            message="version",
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.client.factory.ClientFactory", autospec=True)
    @mock.patch("aprsd.packets.WatchList.is_old")
    def test_callsign_in_watchlist_not_old(self, mock_is_old, mock_factory):
        client.client_factory = mock_factory
        mock_is_old.return_value = False
        self.config_and_init(
            watchlist_enabled=True,
            watchlist_callsigns=["WB4BOR"],
        )
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            fromcall="WB4BOR",
            message="ping",
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.client.factory.ClientFactory", autospec=True)
    @mock.patch("aprsd.packets.WatchList.is_old")
    def test_callsign_in_watchlist_old_same_alert_callsign(self, mock_is_old, mock_factory):
        client.client_factory = mock_factory
        mock_is_old.return_value = True
        self.config_and_init(
            watchlist_enabled=True,
            watchlist_alert_callsign="WB4BOR",
            watchlist_callsigns=["WB4BOR"],
        )
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            fromcall="WB4BOR",
            message="ping",
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.client.factory.ClientFactory", autospec=True)
    @mock.patch("aprsd.packets.WatchList.is_old")
    def test_callsign_in_watchlist_old_send_alert(self, mock_is_old, mock_factory):
        client.client_factory = mock_factory
        mock_is_old.return_value = True
        notify_callsign = fake.FAKE_TO_CALLSIGN
        fromcall = "WB4BOR"
        self.config_and_init(
            watchlist_enabled=True,
            watchlist_alert_callsign=notify_callsign,
            watchlist_callsigns=["WB4BOR"],
        )
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            fromcall=fromcall,
            message="ping",
            msg_number=1,
        )
        packet_type = packet.__class__.__name__
        actual = plugin.filter(packet)
        msg = f"{fromcall} was just seen by type:'{packet_type}'"

        self.assertIsInstance(actual, packets.MessagePacket)
        self.assertEqual(fake.FAKE_FROM_CALLSIGN, actual.from_call)
        self.assertEqual(notify_callsign, actual.to_call)
        self.assertEqual(msg, actual.message_text)



import logging
import shutil
import subprocess

from aprsd import packets, plugin
from aprsd.utils import trace


LOG = logging.getLogger("APRSD")

DEFAULT_FORTUNE_PATH = '/usr/games/fortune'


class FortunePlugin(plugin.APRSDRegexCommandPluginBase):
    """Fortune."""

    command_regex = r"^([f]|[f]\s|fortune)"
    command_name = "fortune"
    short_description = "Give me a fortune"

    fortune_path = None

    def setup(self):
        self.fortune_path = shutil.which(DEFAULT_FORTUNE_PATH)
        LOG.info(f"Fortune path {self.fortune_path}")
        if not self.fortune_path:
            self.enabled = False
        else:
            self.enabled = True

    @trace.trace
    def process(self, packet: packets.MessagePacket):
        LOG.info("FortunePlugin")

        # fromcall = packet.get("from")
        # message = packet.get("message_text", None)
        # ack = packet.get("msgNo", "0")

        reply = None

        try:
            cmnd = [self.fortune_path, "-s", "-n 60"]
            command = " ".join(cmnd)
            output = subprocess.check_output(
                command,
                shell=True,
                timeout=3,
                universal_newlines=True,
            )
            output = (
                output.replace("\r", "")
                .replace("\n", "")
                .replace("  ", "")
                .replace("\t", " ")
            )
        except subprocess.CalledProcessError as ex:
            reply = f"Fortune command failed '{ex.output}'"
        else:
            reply = output

        return reply



import abc
import logging
import threading

from oslo_config import cfg
import wrapt

from aprsd.packets import core


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class APRSClient:
    """Singleton client class that constructs the aprslib connection."""

    _instance = None
    _client = None

    connected = False
    filter = None
    lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        """This magic turns this into a singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # Put any initialization here.
            cls._instance._create_client()
        return cls._instance

    @abc.abstractmethod
    def stats(self) -> dict:
        pass

    def set_filter(self, filter):
        self.filter = filter
        if self._client:
            self._client.set_filter(filter)

    @property
    def client(self):
        if not self._client:
            self._create_client()
        return self._client

    def _create_client(self):
        self._client = self.setup_connection()
        if self.filter:
            LOG.info("Creating APRS client filter")
            self._client.set_filter(self.filter)

    def stop(self):
        if self._client:
            LOG.info("Stopping client connection.")
            self._client.stop()

    def send(self, packet: core.Packet):
        """Send a packet to the network."""
        self.client.send(packet)

    @wrapt.synchronized(lock)
    def reset(self):
        """Call this to force a rebuild/reconnect."""
        LOG.info("Resetting client connection.")
        if self._client:
            self._client.close()
            del self._client
            self._create_client()
        else:
            LOG.warning("Client not initialized, nothing to reset.")

        # Recreate the client
        LOG.info(f"Creating new client {self.client}")

    @abc.abstractmethod
    def setup_connection(self):
        pass

    @staticmethod
    @abc.abstractmethod
    def is_enabled():
        pass

    @staticmethod
    @abc.abstractmethod
    def transport():
        pass

    @abc.abstractmethod
    def decode_packet(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def consumer(self, callback, blocking=False, immortal=False, raw=False):
        pass

    @abc.abstractmethod
    def is_alive(self):
        pass

    @abc.abstractmethod
    def close(self):
        pass



import threading

from oslo_config import cfg
import wrapt

from aprsd import client
from aprsd.utils import singleton


CONF = cfg.CONF


@singleton
class APRSClientStats:

    lock = threading.Lock()

    @wrapt.synchronized(lock)
    def stats(self, serializable=False):
        cl = client.client_factory.create()
        stats = {
            "transport": cl.transport(),
            "filter": cl.filter,
            "connected": cl.connected,
        }

        if cl.transport() == client.TRANSPORT_APRSIS:
            stats["server_string"] = cl.client.server_string
            keepalive = cl.client.aprsd_keepalive
            if serializable:
                keepalive = keepalive.isoformat()
            stats["server_keepalive"] = keepalive
        elif cl.transport() == client.TRANSPORT_TCPKISS:
            stats["host"] = CONF.kiss_tcp.host
            stats["port"] = CONF.kiss_tcp.port
        elif cl.transport() == client.TRANSPORT_SERIALKISS:
            stats["device"] = CONF.kiss_serial.device
        return stats



#
# Listen on amateur radio aprs-is network for messages and respond to them.
# You must have an amateur radio callsign to use this software.  You must
# create an ~/.aprsd/config.yml file with all of the required settings.  To
# generate an example config.yml, just run aprsd, then copy the sample config
# to ~/.aprsd/config.yml and edit the settings.
#
# APRS messages:
#   l(ocation)             = descriptive location of calling station
#   w(eather)              = temp, (hi/low) forecast, later forecast
#   t(ime)                 = respond with the current time
#   f(ortune)              = respond with a short fortune
#   -email_addr email text = send an email
#   -2                     = display the last 2 emails received
#   p(ing)                 = respond with Pong!/time
#   anything else          = respond with usage
#
# (C)2018 Craig Lamparter
# License GPLv2
#

# python included libs
import datetime
import importlib.metadata as imp
from importlib.metadata import version as metadata_version
import logging
import signal
import sys
import time

import click
from oslo_config import cfg, generator

# local imports here
import aprsd
from aprsd import cli_helper, packets, threads, utils
from aprsd.stats import collector


# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
CONF = cfg.CONF
LOG = logging.getLogger("APRSD")
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
flask_enabled = False


@click.group(cls=cli_helper.AliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option()
@click.pass_context
def cli(ctx):
    pass


def load_commands():
    from .cmds import (  # noqa
        completion, dev, fetch_stats, healthcheck, list_plugins, listen,
        send_message, server, webchat,
    )


def main():
    # First import all the possible commands for the CLI
    # The commands themselves live in the cmds directory
    load_commands()
    utils.load_entry_points("aprsd.extension")
    cli(auto_envvar_prefix="APRSD")


def signal_handler(sig, frame):
    global flask_enabled

    click.echo("signal_handler: called")
    threads.APRSDThreadList().stop_all()
    if "subprocess" not in str(frame):
        LOG.info(
            "Ctrl+C, Sending all threads exit! Can take up to 10 seconds {}".format(
                datetime.datetime.now(),
            ),
        )
        time.sleep(1.5)
        packets.PacketTrack().save()
        packets.WatchList().save()
        packets.SeenList().save()
        packets.PacketList().save()
        collector.Collector().collect()
        # signal.signal(signal.SIGTERM, sys.exit(0))
        # sys.exit(0)

    if flask_enabled:
        signal.signal(signal.SIGTERM, sys.exit(0))


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.pass_context
@cli_helper.process_standard_options_no_config
def check_version(ctx):
    """Check this version against the latest in pypi.org."""
    level, msg = utils._check_version()
    if level:
        click.secho(msg, fg="yellow")
    else:
        click.secho(msg, fg="green")


@cli.command()
@click.pass_context
def sample_config(ctx):
    """Generate a sample Config file from aprsd and all installed plugins."""

    def _get_selected_entry_points():
        import sys
        if sys.version_info < (3, 10):
            all = imp.entry_points()
            selected = []
            if "oslo.config.opts" in all:
                for x in all["oslo.config.opts"]:
                    if x.group == "oslo.config.opts":
                        selected.append(x)
        else:
            selected = imp.entry_points(group="oslo.config.opts")

        return selected

    def get_namespaces():
        args = []

        # selected = imp.entry_points(group="oslo.config.opts")
        selected = _get_selected_entry_points()
        for entry in selected:
            if "aprsd" in entry.name:
                args.append("--namespace")
                args.append(entry.name)

        return args

    args = get_namespaces()
    config_version = metadata_version("oslo.config")
    logging.basicConfig(level=logging.WARN)
    conf = cfg.ConfigOpts()
    generator.register_cli_opts(conf)
    try:
        conf(args, version=config_version)
    except cfg.RequiredOptError:
        conf.print_help()
        if not sys.argv[1:]:
            raise SystemExit
        raise
    generator.generate(conf)
    return


@cli.command()
@click.pass_context
def version(ctx):
    """Show the APRSD version."""
    click.echo(click.style("APRSD Version : ", fg="white"), nl=False)
    click.secho(f"{aprsd.__version__}", fg="yellow", bold=True)


if __name__ == "__main__":
    main()



import logging
import re
import time

from geopy.geocoders import ArcGIS, AzureMaps, Baidu, Bing, GoogleV3
from geopy.geocoders import HereV7, Nominatim, OpenCage, TomTom, What3WordsV3, Woosmap
from oslo_config import cfg

from aprsd import packets, plugin, plugin_utils
from aprsd.utils import trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class UsLocation:
    raw = {}

    def __init__(self, info):
        self.info = info

    def __str__(self):
        return self.info


class USGov:
    """US Government geocoder that uses the geopy API.

    This is a dummy class the implements the geopy reverse API,
    so the factory can return an object that conforms to the API.
    """
    def reverse(self, coordinates):
        """Reverse geocode a coordinate."""
        LOG.info(f"USGov reverse geocode {coordinates}")
        coords = coordinates.split(",")
        lat = float(coords[0])
        lon = float(coords[1])
        result = plugin_utils.get_weather_gov_for_gps(lat, lon)
        # LOG.info(f"WEATHER: {result}")
        # LOG.info(f"area description {result['location']['areaDescription']}")
        if 'location' in result:
            loc = UsLocation(result['location']['areaDescription'])
        else:
            loc = UsLocation("Unknown Location")

        LOG.info(f"USGov reverse geocode LOC {loc}")
        return loc


def geopy_factory():
    """Factory function for geopy geocoders."""
    geocoder = CONF.location_plugin.geopy_geocoder
    LOG.info(f"Using geocoder: {geocoder}")
    user_agent = CONF.location_plugin.user_agent
    LOG.info(f"Using user_agent: {user_agent}")

    if geocoder == "Nominatim":
        return Nominatim(user_agent=user_agent)
    elif geocoder == "USGov":
        return USGov()
    elif geocoder == "ArcGIS":
        return ArcGIS(
            username=CONF.location_plugin.arcgis_username,
            password=CONF.location_plugin.arcgis_password,
            user_agent=user_agent,
        )
    elif geocoder == "AzureMaps":
        return AzureMaps(
            user_agent=user_agent,
            subscription_key=CONF.location_plugin.azuremaps_subscription_key,
        )
    elif geocoder == "Baidu":
        return Baidu(user_agent=user_agent, api_key=CONF.location_plugin.baidu_api_key)
    elif geocoder == "Bing":
        return Bing(user_agent=user_agent, api_key=CONF.location_plugin.bing_api_key)
    elif geocoder == "GoogleV3":
        return GoogleV3(user_agent=user_agent, api_key=CONF.location_plugin.google_api_key)
    elif geocoder == "HERE":
        return HereV7(user_agent=user_agent, api_key=CONF.location_plugin.here_api_key)
    elif geocoder == "OpenCage":
        return OpenCage(user_agent=user_agent, api_key=CONF.location_plugin.opencage_api_key)
    elif geocoder == "TomTom":
        return TomTom(user_agent=user_agent, api_key=CONF.location_plugin.tomtom_api_key)
    elif geocoder == "What3Words":
        return What3WordsV3(user_agent=user_agent, api_key=CONF.location_plugin.what3words_api_key)
    elif geocoder == "Woosmap":
        return Woosmap(user_agent=user_agent, api_key=CONF.location_plugin.woosmap_api_key)
    else:
        raise ValueError(f"Unknown geocoder: {geocoder}")


class LocationPlugin(plugin.APRSDRegexCommandPluginBase, plugin.APRSFIKEYMixin):
    """Location!"""

    command_regex = r"^([l]|[l]\s|location)"
    command_name = "location"
    short_description = "Where in the world is a CALLSIGN's last GPS beacon?"

    def setup(self):
        self.ensure_aprs_fi_key()

    @trace.trace
    def process(self, packet: packets.MessagePacket):
        LOG.info("Location Plugin")
        fromcall = packet.from_call
        message = packet.get("message_text", None)

        api_key = CONF.aprs_fi.apiKey

        # optional second argument is a callsign to search
        a = re.search(r"^.*\s+(.*)", message)
        if a is not None:
            searchcall = a.group(1)
            searchcall = searchcall.upper()
        else:
            # if no second argument, search for calling station
            searchcall = fromcall

        try:
            aprs_data = plugin_utils.get_aprs_fi(api_key, searchcall)
        except Exception as ex:
            LOG.error(f"Failed to fetch aprs.fi '{ex}'")
            return "Failed to fetch aprs.fi location"

        LOG.debug(f"LocationPlugin: aprs_data = {aprs_data}")
        if not len(aprs_data["entries"]):
            LOG.error("Didn't get any entries from aprs.fi")
            return "Failed to fetch aprs.fi location"

        lat = float(aprs_data["entries"][0]["lat"])
        lon = float(aprs_data["entries"][0]["lng"])

        # Get some information about their location
        try:
            tic = time.perf_counter()
            geolocator = geopy_factory()
            LOG.info(f"Using GEOLOCATOR: {geolocator}")
            coordinates = f"{lat:0.6f}, {lon:0.6f}"
            location = geolocator.reverse(coordinates)
            address = location.raw.get("address")
            LOG.debug(f"GEOLOCATOR address: {address}")
            toc = time.perf_counter()
            if address:
                LOG.info(f"Geopy address {address} took {toc - tic:0.4f}")
                if address.get("country_code") == "us":
                    area_info = f"{address.get('county')}, {address.get('state')}"
                else:
                    # what to do for address for non US?
                    area_info = f"{address.get('country'), 'Unknown'}"
            else:
                area_info = str(location)
        except Exception as ex:
            LOG.error(ex)
            LOG.error(f"Failed to fetch Geopy address {ex}")
            area_info = "Unknown Location"

        try:  # altitude not always provided
            alt = float(aprs_data["entries"][0]["altitude"])
        except Exception:
            alt = 0
        altfeet = int(alt * 3.28084)
        aprs_lasttime_seconds = aprs_data["entries"][0]["lasttime"]
        # aprs_lasttime_seconds = aprs_lasttime_seconds.encode(
        #    "ascii", errors="ignore"
        # )  # unicode to ascii
        delta_seconds = time.time() - int(aprs_lasttime_seconds)
        delta_hours = delta_seconds / 60 / 60

        reply = "{}: {} {}' {},{} {}h ago".format(
            searchcall,
            area_info,
            str(altfeet),
            f"{lat:0.2f}",
            f"{lon:0.2f}",
            str("%.1f" % round(delta_hours, 1)),
        ).rstrip()

        return reply



import logging
from typing import Callable, Protocol, runtime_checkable

from aprsd import exception
from aprsd.packets import core


LOG = logging.getLogger("APRSD")


@runtime_checkable
class Client(Protocol):

    def __init__(self):
        pass

    def connect(self) -> bool:
        pass

    def disconnect(self) -> bool:
        pass

    def decode_packet(self, *args, **kwargs) -> type[core.Packet]:
        pass

    def is_enabled(self) -> bool:
        pass

    def is_configured(self) -> bool:
        pass

    def transport(self) -> str:
        pass

    def send(self, message: str) -> bool:
        pass

    def setup_connection(self) -> None:
        pass


class ClientFactory:
    _instance = None
    clients = []

    def __new__(cls, *args, **kwargs):
        """This magic turns this into a singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # Put any initialization here.
        return cls._instance

    def __init__(self):
        self.clients: list[Callable] = []

    def register(self, aprsd_client: Callable):
        if isinstance(aprsd_client, Client):
            raise ValueError("Client must be a subclass of Client protocol")

        self.clients.append(aprsd_client)

    def create(self, key=None):
        for client in self.clients:
            if client.is_enabled():
                return client()
        raise Exception("No client is configured!!")

    def is_client_enabled(self):
        """Make sure at least one client is enabled."""
        enabled = False
        for client in self.clients:
            if client.is_enabled():
                enabled = True
        return enabled

    def is_client_configured(self):
        enabled = False
        for client in self.clients:
            try:
                if client.is_configured():
                    enabled = True
            except exception.MissingConfigOptionException as ex:
                LOG.error(ex.message)
                return False
            except exception.ConfigOptionBogusDefaultException as ex:
                LOG.error(ex.message)
                return False
        return enabled



import logging

from oslo_config import cfg

from aprsd import client
from aprsd.client import base
from aprsd.client.drivers import fake as fake_driver
from aprsd.utils import trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class APRSDFakeClient(base.APRSClient, metaclass=trace.TraceWrapperMetaclass):

    def stats(self) -> dict:
        return {}

    @staticmethod
    def is_enabled():
        if CONF.fake_client.enabled:
            return True
        return False

    @staticmethod
    def is_configured():
        return APRSDFakeClient.is_enabled()

    def is_alive(self):
        return True

    def close(self):
        pass

    def setup_connection(self):
        self.connected = True
        return fake_driver.APRSDFakeClient()

    @staticmethod
    def transport():
        return client.TRANSPORT_FAKE

    def decode_packet(self, *args, **kwargs):
        LOG.debug(f"kwargs {kwargs}")
        pkt = kwargs["packet"]
        LOG.debug(f"Got an APRS Fake Packet '{pkt}'")
        return pkt



# Fetch active stats from a remote running instance of aprsd admin web interface.
import logging

import click
from oslo_config import cfg
import requests
from rich.console import Console
from rich.table import Table

# local imports here
import aprsd
from aprsd import cli_helper
from aprsd.main import cli


# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
LOG = logging.getLogger("APRSD")
CONF = cfg.CONF


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "--host", type=str,
    default=None,
    help="IP address of the remote aprsd admin web ui fetch stats from.",
)
@click.option(
    "--port", type=int,
    default=None,
    help="Port of the remote aprsd web admin interface to fetch stats from.",
)
@click.pass_context
@cli_helper.process_standard_options
def fetch_stats(ctx, host, port):
    """Fetch stats from a APRSD admin web interface."""
    console = Console()
    console.print(f"APRSD Fetch-Stats started version: {aprsd.__version__}")

    CONF.log_opt_values(LOG, logging.DEBUG)
    if not host:
        host = CONF.admin.web_ip
    if not port:
        port = CONF.admin.web_port

    msg = f"Fetching stats from {host}:{port}"
    console.print(msg)
    with console.status(msg):
        response = requests.get(f"http://{host}:{port}/stats", timeout=120)
        if not response:
            console.print(
                f"Failed to fetch stats from {host}:{port}?",
                style="bold red",
            )
            return

        stats = response.json()
        if not stats:
            console.print(
                f"Failed to fetch stats from aprsd admin ui at {host}:{port}",
                style="bold red",
            )
            return

    aprsd_title = (
        "APRSD "
        f"[bold cyan]v{stats['APRSDStats']['version']}[/] "
        f"Callsign [bold green]{stats['APRSDStats']['callsign']}[/] "
        f"Uptime [bold yellow]{stats['APRSDStats']['uptime']}[/]"
    )

    console.rule(f"Stats from {host}:{port}")
    console.print("\n\n")
    console.rule(aprsd_title)

    # Show the connection to APRS
    # It can be a connection to an APRS-IS server or a local TNC via KISS or KISSTCP
    if "aprs-is" in stats:
        title = f"APRS-IS Connection {stats['APRSClientStats']['server_string']}"
        table = Table(title=title)
        table.add_column("Key")
        table.add_column("Value")
        for key, value in stats["APRSClientStats"].items():
            table.add_row(key, value)
        console.print(table)

    threads_table = Table(title="Threads")
    threads_table.add_column("Name")
    threads_table.add_column("Alive?")
    for name, alive in stats["APRSDThreadList"].items():
        threads_table.add_row(name, str(alive))

    console.print(threads_table)

    packet_totals = Table(title="Packet Totals")
    packet_totals.add_column("Key")
    packet_totals.add_column("Value")
    packet_totals.add_row("Total Received", str(stats["PacketList"]["rx"]))
    packet_totals.add_row("Total Sent", str(stats["PacketList"]["tx"]))
    console.print(packet_totals)

    # Show each of the packet types
    packets_table = Table(title="Packets By Type")
    packets_table.add_column("Packet Type")
    packets_table.add_column("TX")
    packets_table.add_column("RX")
    for key, value in stats["PacketList"]["packets"].items():
        packets_table.add_row(key, str(value["tx"]), str(value["rx"]))

    console.print(packets_table)

    if "plugins" in stats:
        count = len(stats["PluginManager"])
        plugins_table = Table(title=f"Plugins ({count})")
        plugins_table.add_column("Plugin")
        plugins_table.add_column("Enabled")
        plugins_table.add_column("Version")
        plugins_table.add_column("TX")
        plugins_table.add_column("RX")
        plugins = stats["PluginManager"]
        for key, value in plugins.items():
            plugins_table.add_row(
                key,
                str(plugins[key]["enabled"]),
                plugins[key]["version"],
                str(plugins[key]["tx"]),
                str(plugins[key]["rx"]),
            )

        console.print(plugins_table)

    seen_list = stats.get("SeenList")

    if seen_list:
        count = len(seen_list)
        seen_table = Table(title=f"Seen List ({count})")
        seen_table.add_column("Callsign")
        seen_table.add_column("Message Count")
        seen_table.add_column("Last Heard")
        for key, value in seen_list.items():
            seen_table.add_row(key, str(value["count"]), value["last"])

        console.print(seen_table)

    watch_list = stats.get("WatchList")

    if watch_list:
        count = len(watch_list)
        watch_table = Table(title=f"Watch List ({count})")
        watch_table.add_column("Callsign")
        watch_table.add_column("Last Heard")
        for key, value in watch_list.items():
            watch_table.add_row(key, value["last"])

        console.print(watch_table)



from aprsd.packets import collector
from aprsd.packets.core import (  # noqa: F401
    AckPacket, BeaconPacket, BulletinPacket, GPSPacket, MessagePacket,
    MicEPacket, ObjectPacket, Packet, RejectPacket, StatusPacket,
    ThirdPartyPacket, UnknownPacket, WeatherPacket, factory,
)
from aprsd.packets.packet_list import PacketList  # noqa: F401
from aprsd.packets.seen_list import SeenList  # noqa: F401
from aprsd.packets.tracker import PacketTrack  # noqa: F401
from aprsd.packets.watch_list import WatchList  # noqa: F401


# Register all the packet tracking objects.
collector.PacketCollector().register(PacketList)
collector.PacketCollector().register(SeenList)
collector.PacketCollector().register(PacketTrack)
collector.PacketCollector().register(WatchList)


NULL_MESSAGE = -1



import abc
import logging
import queue
import time

import aprslib
from oslo_config import cfg

from aprsd import packets, plugin
from aprsd.client import client_factory
from aprsd.packets import collector
from aprsd.packets import log as packet_log
from aprsd.threads import APRSDThread, tx


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class APRSDRXThread(APRSDThread):
    def __init__(self, packet_queue):
        super().__init__("RX_PKT")
        self.packet_queue = packet_queue
        self._client = client_factory.create()

    def stop(self):
        self.thread_stop = True
        if self._client:
            self._client.stop()

    def loop(self):
        if not self._client:
            self._client = client_factory.create()
            time.sleep(1)
            return True
        # setup the consumer of messages and block until a messages
        try:
            # This will register a packet consumer with aprslib
            # When new packets come in the consumer will process
            # the packet

            # Do a partial here because the consumer signature doesn't allow
            # For kwargs to be passed in to the consumer func we declare
            # and the aprslib developer didn't want to allow a PR to add
            # kwargs.  :(
            # https://github.com/rossengeorgiev/aprs-python/pull/56
            self._client.consumer(
                self._process_packet, raw=False, blocking=False,
            )
        except (
            aprslib.exceptions.ConnectionDrop,
            aprslib.exceptions.ConnectionError,
        ):
            LOG.error("Connection dropped, reconnecting")
            # Force the deletion of the client object connected to aprs
            # This will cause a reconnect, next time client.get_client()
            # is called
            self._client.reset()
            time.sleep(5)
        except Exception:
            # LOG.exception(ex)
            LOG.error("Resetting connection and trying again.")
            self._client.reset()
            time.sleep(5)
        # Continue to loop
        return True

    def _process_packet(self, *args, **kwargs):
        """Intermediate callback so we can update the keepalive time."""
        # Now call the 'real' packet processing for a RX'x packet
        self.process_packet(*args, **kwargs)

    @abc.abstractmethod
    def process_packet(self, *args, **kwargs):
        pass


class APRSDDupeRXThread(APRSDRXThread):
    """Process received packets.

    This is the main APRSD Server command thread that
    receives packets and makes sure the packet
    hasn't been seen previously before sending it on
    to be processed.
    """

    def process_packet(self, *args, **kwargs):
        """This handles the processing of an inbound packet.

        When a packet is received by the connected client object,
        it sends the raw packet into this function.  This function then
        decodes the packet via the client, and then processes the packet.
        Ack Packets are sent to the PluginProcessPacketThread for processing.
        All other packets have to be checked as a dupe, and then only after
        we haven't seen this packet before, do we send it to the
        PluginProcessPacketThread for processing.
        """
        packet = self._client.decode_packet(*args, **kwargs)
        # LOG.debug(raw)
        packet_log.log(packet)
        pkt_list = packets.PacketList()

        if isinstance(packet, packets.AckPacket):
            # We don't need to drop AckPackets, those should be
            # processed.
            self.packet_queue.put(packet)
        else:
            # Make sure we aren't re-processing the same packet
            # For RF based APRS Clients we can get duplicate packets
            # So we need to track them and not process the dupes.
            found = False
            try:
                # Find the packet in the list of already seen packets
                # Based on the packet.key
                found = pkt_list.find(packet)
            except KeyError:
                found = False

            if not found:
                # We haven't seen this packet before, so we process it.
                collector.PacketCollector().rx(packet)
                self.packet_queue.put(packet)
            elif packet.timestamp - found.timestamp < CONF.packet_dupe_timeout:
                # If the packet came in within N seconds of the
                # Last time seeing the packet, then we drop it as a dupe.
                LOG.warning(f"Packet {packet.from_call}:{packet.msgNo} already tracked, dropping.")
            else:
                LOG.warning(
                    f"Packet {packet.from_call}:{packet.msgNo} already tracked "
                    f"but older than {CONF.packet_dupe_timeout} seconds. processing.",
                )
                collector.PacketCollector().rx(packet)
                self.packet_queue.put(packet)


class APRSDPluginRXThread(APRSDDupeRXThread):
    """"Process received packets.

    For backwards compatibility, we keep the APRSDPluginRXThread.
    """


class APRSDProcessPacketThread(APRSDThread):
    """Base class for processing received packets.

    This is the base class for processing packets coming from
    the consumer.  This base class handles sending ack packets and
    will ack a message before sending the packet to the subclass
    for processing."""

    def __init__(self, packet_queue):
        self.packet_queue = packet_queue
        super().__init__("ProcessPKT")

    def process_ack_packet(self, packet):
        """We got an ack for a message, no need to resend it."""
        ack_num = packet.msgNo
        LOG.debug(f"Got ack for message {ack_num}")
        collector.PacketCollector().rx(packet)

    def process_piggyback_ack(self, packet):
        """We got an ack embedded in a packet."""
        ack_num = packet.ackMsgNo
        LOG.debug(f"Got PiggyBackAck for message {ack_num}")
        collector.PacketCollector().rx(packet)

    def process_reject_packet(self, packet):
        """We got a reject message for a packet.  Stop sending the message."""
        ack_num = packet.msgNo
        LOG.debug(f"Got REJECT for message {ack_num}")
        collector.PacketCollector().rx(packet)

    def loop(self):
        try:
            packet = self.packet_queue.get(timeout=1)
            if packet:
                self.process_packet(packet)
        except queue.Empty:
            pass
        return True

    def process_packet(self, packet):
        """Process a packet received from aprs-is server."""
        LOG.debug(f"ProcessPKT-LOOP {self.loop_count}")
        our_call = CONF.callsign.lower()

        from_call = packet.from_call
        if packet.addresse:
            to_call = packet.addresse
        else:
            to_call = packet.to_call
        msg_id = packet.msgNo

        # We don't put ack packets destined for us through the
        # plugins.
        if (
            isinstance(packet, packets.AckPacket)
            and packet.addresse.lower() == our_call
        ):
            self.process_ack_packet(packet)
        elif (
            isinstance(packet, packets.RejectPacket)
            and packet.addresse.lower() == our_call
        ):
            self.process_reject_packet(packet)
        else:
            if hasattr(packet, "ackMsgNo") and packet.ackMsgNo:
                # we got an ack embedded in this packet
                # we need to handle the ack
                self.process_piggyback_ack(packet)
            # Only ack messages that were sent directly to us
            if isinstance(packet, packets.MessagePacket):
                if to_call and to_call.lower() == our_call:
                    # It's a MessagePacket and it's for us!
                    # let any threads do their thing, then ack
                    # send an ack last
                    tx.send(
                        packets.AckPacket(
                            from_call=CONF.callsign,
                            to_call=from_call,
                            msgNo=msg_id,
                        ),
                    )

                    self.process_our_message_packet(packet)
                else:
                    # Packet wasn't meant for us!
                    self.process_other_packet(packet, for_us=False)
            else:
                self.process_other_packet(
                    packet, for_us=(to_call.lower() == our_call),
                )
        LOG.debug(f"Packet processing complete for pkt '{packet.key}'")
        return False

    @abc.abstractmethod
    def process_our_message_packet(self, packet):
        """Process a MessagePacket destined for us!"""

    def process_other_packet(self, packet, for_us=False):
        """Process an APRS Packet that isn't a message or ack"""
        if not for_us:
            LOG.info("Got a packet not meant for us.")
        else:
            LOG.info("Got a non AckPacket/MessagePacket")


class APRSDPluginProcessPacketThread(APRSDProcessPacketThread):
    """Process the packet through the plugin manager.

    This is the main aprsd server plugin processing thread."""

    def process_other_packet(self, packet, for_us=False):
        pm = plugin.PluginManager()
        try:
            results = pm.run_watchlist(packet)
            for reply in results:
                if isinstance(reply, list):
                    for subreply in reply:
                        LOG.debug(f"Sending '{subreply}'")
                        if isinstance(subreply, packets.Packet):
                            tx.send(subreply)
                        else:
                            wl = CONF.watch_list
                            to_call = wl["alert_callsign"]
                            tx.send(
                                packets.MessagePacket(
                                    from_call=CONF.callsign,
                                    to_call=to_call,
                                    message_text=subreply,
                                ),
                            )
                elif isinstance(reply, packets.Packet):
                    # We have a message based object.
                    tx.send(reply)
        except Exception as ex:
            LOG.error("Plugin failed!!!")
            LOG.exception(ex)

    def process_our_message_packet(self, packet):
        """Send the packet through the plugins."""
        from_call = packet.from_call
        if packet.addresse:
            to_call = packet.addresse
        else:
            to_call = None

        pm = plugin.PluginManager()
        try:
            results = pm.run(packet)
            replied = False
            for reply in results:
                if isinstance(reply, list):
                    # one of the plugins wants to send multiple messages
                    replied = True
                    for subreply in reply:
                        LOG.debug(f"Sending '{subreply}'")
                        if isinstance(subreply, packets.Packet):
                            tx.send(subreply)
                        else:
                            tx.send(
                                packets.MessagePacket(
                                    from_call=CONF.callsign,
                                    to_call=from_call,
                                    message_text=subreply,
                                ),
                            )
                elif isinstance(reply, packets.Packet):
                    # We have a message based object.
                    tx.send(reply)
                    replied = True
                else:
                    replied = True
                    # A plugin can return a null message flag which signals
                    # us that they processed the message correctly, but have
                    # nothing to reply with, so we avoid replying with a
                    # usage string
                    if reply is not packets.NULL_MESSAGE:
                        LOG.debug(f"Sending '{reply}'")
                        tx.send(
                            packets.MessagePacket(
                                from_call=CONF.callsign,
                                to_call=from_call,
                                message_text=reply,
                            ),
                        )

            # If the message was for us and we didn't have a
            # response, then we send a usage statement.
            if to_call == CONF.callsign and not replied:
                LOG.warning("Sending help!")
                message_text = "Unknown command! Send 'help' message for help"
                tx.send(
                    packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=from_call,
                        message_text=message_text,
                    ),
                )
        except Exception as ex:
            LOG.error("Plugin failed!!!")
            LOG.exception(ex)
            # Do we need to send a reply?
            if to_call == CONF.callsign:
                reply = "A Plugin failed! try again?"
                tx.send(
                    packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=from_call,
                        message_text=reply,
                    ),
                )

        LOG.debug("Completed process_our_message_packet")



import logging
import threading
import time

import aprslib
from oslo_config import cfg
import wrapt

from aprsd import conf  # noqa
from aprsd.packets import core
from aprsd.utils import trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class APRSDFakeClient(metaclass=trace.TraceWrapperMetaclass):
    '''Fake client for testing.'''

    # flag to tell us to stop
    thread_stop = False

    lock = threading.Lock()
    path = []

    def __init__(self):
        LOG.info("Starting APRSDFakeClient client.")
        self.path = ["WIDE1-1", "WIDE2-1"]

    def stop(self):
        self.thread_stop = True
        LOG.info("Shutdown APRSDFakeClient client.")

    def is_alive(self):
        """If the connection is alive or not."""
        return not self.thread_stop

    @wrapt.synchronized(lock)
    def send(self, packet: core.Packet):
        """Send an APRS Message object."""
        LOG.info(f"Sending packet: {packet}")
        payload = None
        if isinstance(packet, core.Packet):
            packet.prepare()
            payload = packet.payload.encode("US-ASCII")
            if packet.path:
                packet.path
            else:
                self.path
        else:
            msg_payload = f"{packet.raw}{{{str(packet.msgNo)}"
            payload = (
                ":{:<9}:{}".format(
                    packet.to_call,
                    msg_payload,
                )
            ).encode("US-ASCII")

        LOG.debug(
            f"FAKE::Send '{payload}' TO '{packet.to_call}' From "
            f"'{packet.from_call}' with PATH \"{self.path}\"",
        )

    def consumer(self, callback, blocking=False, immortal=False, raw=False):
        LOG.debug("Start non blocking FAKE consumer")
        # Generate packets here?
        raw = "GTOWN>APDW16,WIDE1-1,WIDE2-1:}KM6LYW-9>APZ100,TCPIP,GTOWN*::KM6LYW   :KM6LYW: 19 Miles SW"
        pkt_raw = aprslib.parse(raw)
        pkt = core.factory(pkt_raw)
        callback(packet=pkt)
        LOG.debug(f"END blocking FAKE consumer {self}")
        time.sleep(8)



from pathlib import Path

from oslo_config import cfg


home = str(Path.home())
DEFAULT_CONFIG_DIR = f"{home}/.config/aprsd/"
APRSD_DEFAULT_MAGIC_WORD = "CHANGEME!!!"

admin_group = cfg.OptGroup(
    name="admin",
    title="Admin web interface settings",
)
watch_list_group = cfg.OptGroup(
    name="watch_list",
    title="Watch List settings",
)
webchat_group = cfg.OptGroup(
    name="webchat",
    title="Settings specific to the webchat command",
)

registry_group = cfg.OptGroup(
    name="aprs_registry",
    title="APRS Registry settings",
)


aprsd_opts = [
    cfg.StrOpt(
        "callsign",
        required=True,
        help="Callsign to use for messages sent by APRSD",
    ),
    cfg.BoolOpt(
        "enable_save",
        default=True,
        help="Enable saving of watch list, packet tracker between restarts.",
    ),
    cfg.StrOpt(
        "save_location",
        default=DEFAULT_CONFIG_DIR,
        help="Save location for packet tracking files.",
    ),
    cfg.BoolOpt(
        "trace_enabled",
        default=False,
        help="Enable code tracing",
    ),
    cfg.StrOpt(
        "units",
        default="imperial",
        help="Units for display, imperial or metric",
    ),
    cfg.IntOpt(
        "ack_rate_limit_period",
        default=1,
        help="The wait period in seconds per Ack packet being sent."
             "1 means 1 ack packet per second allowed."
             "2 means 1 pack packet every 2 seconds allowed",
    ),
    cfg.IntOpt(
        "msg_rate_limit_period",
        default=2,
        help="Wait period in seconds per non AckPacket being sent."
             "2 means 1 packet every 2 seconds allowed."
             "5 means 1 pack packet every 5 seconds allowed",
    ),
    cfg.IntOpt(
        "packet_dupe_timeout",
        default=300,
        help="The number of seconds before a packet is not considered a duplicate.",
    ),
    cfg.BoolOpt(
        "enable_beacon",
        default=False,
        help="Enable sending of a GPS Beacon packet to locate this service. "
             "Requires latitude and longitude to be set.",
    ),
    cfg.IntOpt(
        "beacon_interval",
        default=1800,
        help="The number of seconds between beacon packets.",
    ),
    cfg.StrOpt(
        "beacon_symbol",
        default="/",
        help="The symbol to use for the GPS Beacon packet. See: http://www.aprs.net/vm/DOS/SYMBOLS.HTM",
    ),
    cfg.StrOpt(
        "latitude",
        default=None,
        help="Latitude for the GPS Beacon button.  If not set, the button will not be enabled.",
    ),
    cfg.StrOpt(
        "longitude",
        default=None,
        help="Longitude for the GPS Beacon button.  If not set, the button will not be enabled.",
    ),
    cfg.StrOpt(
        "log_packet_format",
        choices=["compact", "multiline", "both"],
        default="compact",
        help="When logging packets 'compact' will use a single line formatted for each packet."
             "'multiline' will use multiple lines for each packet and is the traditional format."
             "both will log both compact and multiline.",
    ),
    cfg.IntOpt(
        "default_packet_send_count",
        default=3,
        help="The number of times to send a non ack packet before giving up.",
    ),
    cfg.IntOpt(
        "default_ack_send_count",
        default=3,
        help="The number of times to send an ack packet in response to recieving a packet.",
    ),
    cfg.IntOpt(
        "packet_list_maxlen",
        default=100,
        help="The maximum number of packets to store in the packet list.",
    ),
    cfg.IntOpt(
        "packet_list_stats_maxlen",
        default=20,
        help="The maximum number of packets to send in the stats dict for admin ui.",
    ),
    cfg.BoolOpt(
        "enable_seen_list",
        default=True,
        help="Enable the Callsign seen list tracking feature.  This allows aprsd to keep track of "
             "callsigns that have been seen and when they were last seen.",
    ),
    cfg.BoolOpt(
        "enable_packet_logging",
        default=True,
        help="Set this to False, to disable logging of packets to the log file.",
    ),
]

watch_list_opts = [
    cfg.BoolOpt(
        "enabled",
        default=False,
        help="Enable the watch list feature.  Still have to enable "
             "the correct plugin.  Built-in plugin to use is "
             "aprsd.plugins.notify.NotifyPlugin",
    ),
    cfg.ListOpt(
        "callsigns",
        help="Callsigns to watch for messsages",
    ),
    cfg.StrOpt(
        "alert_callsign",
        help="The Ham Callsign to send messages to for watch list alerts.",
    ),
    cfg.IntOpt(
        "packet_keep_count",
        default=10,
        help="The number of packets to store.",
    ),
    cfg.IntOpt(
        "alert_time_seconds",
        default=3600,
        help="Time to wait before alert is sent on new message for "
             "users in callsigns.",
    ),
]

admin_opts = [
    cfg.BoolOpt(
        "web_enabled",
        default=False,
        help="Enable the Admin Web Interface",
    ),
    cfg.StrOpt(
        "web_ip",
        default="0.0.0.0",
        help="The ip address to listen on",
    ),
    cfg.PortOpt(
        "web_port",
        default=8001,
        help="The port to listen on",
    ),
    cfg.StrOpt(
        "user",
        default="admin",
        help="The admin user for the admin web interface",
    ),
    cfg.StrOpt(
        "password",
        default="password",
        secret=True,
        help="Admin interface password",
    ),
]


enabled_plugins_opts = [
    cfg.ListOpt(
        "enabled_plugins",
        default=[
            "aprsd.plugins.email.EmailPlugin",
            "aprsd.plugins.fortune.FortunePlugin",
            "aprsd.plugins.location.LocationPlugin",
            "aprsd.plugins.ping.PingPlugin",
            "aprsd.plugins.time.TimePlugin",
            "aprsd.plugins.weather.OWMWeatherPlugin",
            "aprsd.plugins.version.VersionPlugin",
            "aprsd.plugins.notify.NotifySeenPlugin",
        ],
        help="Comma separated list of enabled plugins for APRSD."
             "To enable installed external plugins add them here."
             "The full python path to the class name must be used",
    ),
]

webchat_opts = [
    cfg.StrOpt(
        "web_ip",
        default="0.0.0.0",
        help="The ip address to listen on",
    ),
    cfg.PortOpt(
        "web_port",
        default=8001,
        help="The port to listen on",
    ),
    cfg.StrOpt(
        "latitude",
        default=None,
        help="Latitude for the GPS Beacon button.  If not set, the button will not be enabled.",
    ),
    cfg.StrOpt(
        "longitude",
        default=None,
        help="Longitude for the GPS Beacon button.  If not set, the button will not be enabled.",
    ),
    cfg.BoolOpt(
        "disable_url_request_logging",
        default=False,
        help="Disable the logging of url requests in the webchat command.",
    ),
]

registry_opts = [
    cfg.BoolOpt(
        "enabled",
        default=False,
        help="Enable sending aprs registry information.  This will let the "
             "APRS registry know about your service and it's uptime.  "
             "No personal information is sent, just the callsign, uptime and description. "
             "The service callsign is the callsign set in [DEFAULT] section.",
    ),
    cfg.StrOpt(
        "description",
        default=None,
        help="Description of the service to send to the APRS registry. "
             "This is what will show up in the APRS registry."
             "If not set, the description will be the same as the callsign.",
    ),
    cfg.StrOpt(
        "registry_url",
        default="https://aprs.hemna.com/api/v1/registry",
        help="The APRS registry domain name to send the information to.",
    ),
    cfg.StrOpt(
        "service_website",
        default=None,
        help="The website for your APRS service to send to the APRS registry.",
    ),
    cfg.IntOpt(
        "frequency_seconds",
        default=3600,
        help="The frequency in seconds to send the APRS registry information.",
    ),
]


def register_opts(config):
    config.register_opts(aprsd_opts)
    config.register_opts(enabled_plugins_opts)
    config.register_group(admin_group)
    config.register_opts(admin_opts, group=admin_group)
    config.register_group(watch_list_group)
    config.register_opts(watch_list_opts, group=watch_list_group)
    config.register_group(webchat_group)
    config.register_opts(webchat_opts, group=webchat_group)
    config.register_group(registry_group)
    config.register_opts(registry_opts, group=registry_group)


def list_opts():
    return {
        "DEFAULT": (aprsd_opts + enabled_plugins_opts),
        admin_group.name: admin_opts,
        watch_list_group.name: watch_list_opts,
        webchat_group.name: webchat_opts,
        registry_group.name: registry_opts,
    }



from aprsd.client import aprsis, factory, fake, kiss


TRANSPORT_APRSIS = "aprsis"
TRANSPORT_TCPKISS = "tcpkiss"
TRANSPORT_SERIALKISS = "serialkiss"
TRANSPORT_FAKE = "fake"


client_factory = factory.ClientFactory()
client_factory.register(aprsis.APRSISClient)
client_factory.register(kiss.KISSClient)
client_factory.register(fake.APRSDFakeClient)



from oslo_config import cfg


aprsfi_group = cfg.OptGroup(
    name="aprs_fi",
    title="APRS.FI website settings",
)
query_group = cfg.OptGroup(
    name="query_plugin",
    title="Options for the Query Plugin",
)
avwx_group = cfg.OptGroup(
    name="avwx_plugin",
    title="Options for the AVWXWeatherPlugin",
)
owm_wx_group = cfg.OptGroup(
    name="owm_weather_plugin",
    title="Options for the OWMWeatherPlugin",
)

location_group = cfg.OptGroup(
    name="location_plugin",
    title="Options for the LocationPlugin",
)

aprsfi_opts = [
    cfg.StrOpt(
        "apiKey",
        help="Get the apiKey from your aprs.fi account here:"
             "http://aprs.fi/account",
    ),
]

owm_wx_opts = [
    cfg.StrOpt(
        "apiKey",
        help="OWMWeatherPlugin api key to OpenWeatherMap's API."
             "This plugin uses the openweathermap API to fetch"
             "location and weather information."
             "To use this plugin you need to get an openweathermap"
             "account and apikey."
             "https://home.openweathermap.org/api_keys",
    ),
]

avwx_opts = [
    cfg.StrOpt(
        "apiKey",
        help="avwx-api is an opensource project that has"
             "a hosted service here: https://avwx.rest/"
             "You can launch your own avwx-api in a container"
             "by cloning the githug repo here:"
             "https://github.com/avwx-rest/AVWX-API",
    ),
    cfg.StrOpt(
        "base_url",
        default="https://avwx.rest",
        help="The base url for the avwx API.  If you are hosting your own"
             "Here is where you change the url to point to yours.",
    ),
]

location_opts = [
    cfg.StrOpt(
        "geopy_geocoder",
        choices=[
            "ArcGIS", "AzureMaps", "Baidu", "Bing", "GoogleV3", "HERE",
            "Nominatim", "OpenCage", "TomTom", "USGov", "What3Words", "Woosmap",
        ],
        default="Nominatim",
        help="The geopy geocoder to use.  Default is Nominatim."
             "See https://geopy.readthedocs.io/en/stable/#module-geopy.geocoders"
             "for more information.",
    ),
    cfg.StrOpt(
        "user_agent",
        default="APRSD",
        help="The user agent to use for the Nominatim geocoder."
             "See https://geopy.readthedocs.io/en/stable/#module-geopy.geocoders"
             "for more information.",
    ),
    cfg.StrOpt(
        "arcgis_username",
        default=None,
        help="The username to use for the ArcGIS geocoder."
             "See https://geopy.readthedocs.io/en/latest/#arcgis"
             "for more information."
             "Only used for the ArcGIS geocoder.",
    ),
    cfg.StrOpt(
        "arcgis_password",
        default=None,
        help="The password to use for the ArcGIS geocoder."
             "See https://geopy.readthedocs.io/en/latest/#arcgis"
             "for more information."
             "Only used for the ArcGIS geocoder.",
    ),
    cfg.StrOpt(
        "azuremaps_subscription_key",
        help="The subscription key to use for the AzureMaps geocoder."
             "See https://geopy.readthedocs.io/en/latest/#azuremaps"
             "for more information."
             "Only used for the AzureMaps geocoder.",
    ),
    cfg.StrOpt(
        "baidu_api_key",
        help="The API key to use for the Baidu geocoder."
             "See https://geopy.readthedocs.io/en/latest/#baidu"
             "for more information."
             "Only used for the Baidu geocoder.",
    ),
    cfg.StrOpt(
        "bing_api_key",
        help="The API key to use for the Bing geocoder."
             "See https://geopy.readthedocs.io/en/latest/#bing"
             "for more information."
             "Only used for the Bing geocoder.",
    ),
    cfg.StrOpt(
        "google_api_key",
        help="The API key to use for the Google geocoder."
             "See https://geopy.readthedocs.io/en/latest/#googlev3"
             "for more information."
             "Only used for the Google geocoder.",
    ),
    cfg.StrOpt(
        "here_api_key",
        help="The API key to use for the HERE geocoder."
             "See https://geopy.readthedocs.io/en/latest/#here"
             "for more information."
             "Only used for the HERE geocoder.",
    ),
    cfg.StrOpt(
        "opencage_api_key",
        help="The API key to use for the OpenCage geocoder."
             "See https://geopy.readthedocs.io/en/latest/#opencage"
             "for more information."
             "Only used for the OpenCage geocoder.",
    ),
    cfg.StrOpt(
        "tomtom_api_key",
        help="The API key to use for the TomTom geocoder."
             "See https://geopy.readthedocs.io/en/latest/#tomtom"
             "for more information."
             "Only used for the TomTom geocoder.",
    ),
    cfg.StrOpt(
        "what3words_api_key",
        help="The API key to use for the What3Words geocoder."
             "See https://geopy.readthedocs.io/en/latest/#what3words"
             "for more information."
             "Only used for the What3Words geocoder.",
    ),
    cfg.StrOpt(
        "woosmap_api_key",
        help="The API key to use for the Woosmap geocoder."
             "See https://geopy.readthedocs.io/en/latest/#woosmap"
             "for more information."
             "Only used for the Woosmap geocoder.",
    ),
]


def register_opts(config):
    config.register_group(aprsfi_group)
    config.register_opts(aprsfi_opts, group=aprsfi_group)
    config.register_group(query_group)
    config.register_group(owm_wx_group)
    config.register_opts(owm_wx_opts, group=owm_wx_group)
    config.register_group(avwx_group)
    config.register_opts(avwx_opts, group=avwx_group)
    config.register_group(location_group)
    config.register_opts(location_opts, group=location_group)


def list_opts():
    return {
        aprsfi_group.name: aprsfi_opts,
        owm_wx_group.name: owm_wx_opts,
        avwx_group.name: avwx_opts,
        location_group.name: location_opts,
    }



import logging
import signal
import sys

import click
from oslo_config import cfg

import aprsd
from aprsd import cli_helper
from aprsd import main as aprsd_main
from aprsd import plugin, threads, utils
from aprsd.client import client_factory
from aprsd.main import cli
from aprsd.packets import collector as packet_collector
from aprsd.packets import seen_list
from aprsd.threads import keep_alive, log_monitor, registry, rx
from aprsd.threads import stats as stats_thread
from aprsd.threads import tx


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


# main() ###
@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "-f",
    "--flush",
    "flush",
    is_flag=True,
    show_default=True,
    default=False,
    help="Flush out all old aged messages on disk.",
)
@click.pass_context
@cli_helper.process_standard_options
def server(ctx, flush):
    """Start the aprsd server gateway process."""
    signal.signal(signal.SIGINT, aprsd_main.signal_handler)
    signal.signal(signal.SIGTERM, aprsd_main.signal_handler)

    level, msg = utils._check_version()
    if level:
        LOG.warning(msg)
    else:
        LOG.info(msg)
    LOG.info(f"APRSD Started version: {aprsd.__version__}")

    # Initialize the client factory and create
    # The correct client object ready for use
    if not client_factory.is_client_enabled():
        LOG.error("No Clients are enabled in config.")
        sys.exit(-1)

    # Creates the client object
    LOG.info("Creating client connection")
    aprs_client = client_factory.create()
    LOG.info(aprs_client)

    # Create the initial PM singleton and Register plugins
    # We register plugins first here so we can register each
    # plugins config options, so we can dump them all in the
    # log file output.
    LOG.info("Loading Plugin Manager and registering plugins")
    plugin_manager = plugin.PluginManager()
    plugin_manager.setup_plugins()

    # Dump all the config options now.
    CONF.log_opt_values(LOG, logging.DEBUG)
    message_plugins = plugin_manager.get_message_plugins()
    watchlist_plugins = plugin_manager.get_watchlist_plugins()
    LOG.info("Message Plugins enabled and running:")
    for p in message_plugins:
        LOG.info(p)
    LOG.info("Watchlist Plugins enabled and running:")
    for p in watchlist_plugins:
        LOG.info(p)

    # Make sure we have 1 client transport enabled
    if not client_factory.is_client_enabled():
        LOG.error("No Clients are enabled in config.")
        sys.exit(-1)

    if not client_factory.is_client_configured():
        LOG.error("APRS client is not properly configured in config file.")
        sys.exit(-1)

    if not CONF.enable_seen_list:
        # just deregister the class from the packet collector
        packet_collector.PacketCollector().unregister(seen_list.SeenList)

    # Now load the msgTrack from disk if any
    if flush:
        LOG.debug("Flushing All packet tracking objects.")
        packet_collector.PacketCollector().flush()
    else:
        # Try and load saved MsgTrack list
        LOG.debug("Loading saved packet tracking data.")
        packet_collector.PacketCollector().load()

    # Now start all the main processing threads.

    keepalive = keep_alive.KeepAliveThread()
    keepalive.start()

    stats_store_thread = stats_thread.APRSDStatsStoreThread()
    stats_store_thread.start()

    rx_thread = rx.APRSDPluginRXThread(
        packet_queue=threads.packet_queue,
    )
    process_thread = rx.APRSDPluginProcessPacketThread(
        packet_queue=threads.packet_queue,
    )
    rx_thread.start()
    process_thread.start()

    if CONF.enable_beacon:
        LOG.info("Beacon Enabled.  Starting Beacon thread.")
        bcn_thread = tx.BeaconSendThread()
        bcn_thread.start()

    if CONF.aprs_registry.enabled:
        LOG.info("Registry Enabled.  Starting Registry thread.")
        registry_thread = registry.APRSRegistryThread()
        registry_thread.start()

    if CONF.admin.web_enabled:
        log_monitor_thread = log_monitor.LogMonitorThread()
        log_monitor_thread.start()

    rx_thread.join()
    process_thread.join()

    return 0



from collections import OrderedDict
import logging

from oslo_config import cfg

from aprsd.packets import core
from aprsd.utils import objectstore


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class PacketList(objectstore.ObjectStoreMixin):
    """Class to keep track of the packets we tx/rx."""
    _instance = None
    _total_rx: int = 0
    _total_tx: int = 0
    maxlen: int = 100

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.maxlen = CONF.packet_list_maxlen
            cls._instance._init_data()
        return cls._instance

    def _init_data(self):
        self.data = {
            "types": {},
            "packets": OrderedDict(),
        }

    def rx(self, packet: type[core.Packet]):
        """Add a packet that was received."""
        with self.lock:
            self._total_rx += 1
            self._add(packet)
            ptype = packet.__class__.__name__
            if ptype not in self.data["types"]:
                self.data["types"][ptype] = {"tx": 0, "rx": 0}
            self.data["types"][ptype]["rx"] += 1

    def tx(self, packet: type[core.Packet]):
        """Add a packet that was received."""
        with self.lock:
            self._total_tx += 1
            self._add(packet)
            ptype = packet.__class__.__name__
            if ptype not in self.data["types"]:
                self.data["types"][ptype] = {"tx": 0, "rx": 0}
            self.data["types"][ptype]["tx"] += 1

    def add(self, packet):
        with self.lock:
            self._add(packet)

    def _add(self, packet):
        if not self.data.get("packets"):
            self._init_data()
        if packet.key in self.data["packets"]:
            self.data["packets"].move_to_end(packet.key)
        elif len(self.data["packets"]) == self.maxlen:
            self.data["packets"].popitem(last=False)
        self.data["packets"][packet.key] = packet

    def find(self, packet):
        with self.lock:
            return self.data["packets"][packet.key]

    def __len__(self):
        with self.lock:
            return len(self.data["packets"])

    def total_rx(self):
        with self.lock:
            return self._total_rx

    def total_tx(self):
        with self.lock:
            return self._total_tx

    def stats(self, serializable=False) -> dict:
        # limit the number of packets to return to 50
        with self.lock:
            tmp = OrderedDict(
                reversed(
                    list(
                        self.data.get("packets", OrderedDict()).items(),
                    ),
                ),
            )
            pkts = []
            count = 1
            for packet in tmp:
                pkts.append(tmp[packet])
                count += 1
                if count > CONF.packet_list_stats_maxlen:
                    break

            stats = {
                "total_tracked": self._total_rx + self._total_rx,
                "rx": self._total_rx,
                "tx": self._total_tx,
                "types": self.data.get("types", []),
                "packet_count": len(self.data.get("packets", [])),
                "maxlen": self.maxlen,
                "packets": pkts,
            }
            return stats



import datetime
import logging
import threading

from oslo_config import cfg
import requests
import wrapt

from aprsd import threads
from aprsd.log import log


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


def send_log_entries(force=False):
    """Send all of the log entries to the web interface."""
    if CONF.admin.web_enabled:
        if force or LogEntries().is_purge_ready():
            entries = LogEntries().get_all_and_purge()
            if entries:
                try:
                    requests.post(
                        f"http://{CONF.admin.web_ip}:{CONF.admin.web_port}/log_entries",
                        json=entries,
                        auth=(CONF.admin.user, CONF.admin.password),
                    )
                except Exception:
                    LOG.warning(f"Failed to send log entries. len={len(entries)}")


class LogEntries:
    entries = []
    lock = threading.Lock()
    _instance = None
    last_purge = datetime.datetime.now()
    max_delta = datetime.timedelta(
        hours=0.0, minutes=0, seconds=2,
    )

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def stats(self) -> dict:
        return {
            "log_entries": self.entries,
        }

    @wrapt.synchronized(lock)
    def add(self, entry):
        self.entries.append(entry)

    @wrapt.synchronized(lock)
    def get_all_and_purge(self):
        entries = self.entries.copy()
        self.entries = []
        self.last_purge = datetime.datetime.now()
        return entries

    def is_purge_ready(self):
        now = datetime.datetime.now()
        if (
            now - self.last_purge > self.max_delta
            and len(self.entries) > 1
        ):
            return True
        return False

    @wrapt.synchronized(lock)
    def __len__(self):
        return len(self.entries)


class LogMonitorThread(threads.APRSDThread):

    def __init__(self):
        super().__init__("LogMonitorThread")

    def stop(self):
        send_log_entries(force=True)
        super().stop()

    def loop(self):
        try:
            record = log.logging_queue.get(block=True, timeout=2)
            if isinstance(record, list):
                for item in record:
                    entry = self.json_record(item)
                    LogEntries().add(entry)
            else:
                entry = self.json_record(record)
                LogEntries().add(entry)
        except Exception:
            # Just ignore thi
            pass

        send_log_entries()
        return True

    def json_record(self, record):
        entry = {}
        entry["filename"] = record.filename
        entry["funcName"] = record.funcName
        entry["levelname"] = record.levelname
        entry["lineno"] = record.lineno
        entry["module"] = record.module
        entry["name"] = record.name
        entry["pathname"] = record.pathname
        entry["process"] = record.process
        entry["processName"] = record.processName
        if hasattr(record, "stack_info"):
            entry["stack_info"] = record.stack_info
        else:
            entry["stack_info"] = None
        entry["thread"] = record.thread
        entry["threadName"] = record.threadName
        entry["message"] = record.getMessage()
        return entry



import logging

import aprslib
from oslo_config import cfg

from aprsd import client, exception
from aprsd.client import base
from aprsd.client.drivers import kiss
from aprsd.packets import core


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class KISSClient(base.APRSClient):

    _client = None

    def stats(self) -> dict:
        stats = {}
        if self.is_configured():
            return {
                "transport": self.transport(),
            }
        return stats

    @staticmethod
    def is_enabled():
        """Return if tcp or serial KISS is enabled."""
        if CONF.kiss_serial.enabled:
            return True

        if CONF.kiss_tcp.enabled:
            return True

        return False

    @staticmethod
    def is_configured():
        # Ensure that the config vars are correctly set
        if KISSClient.is_enabled():
            transport = KISSClient.transport()
            if transport == client.TRANSPORT_SERIALKISS:
                if not CONF.kiss_serial.device:
                    LOG.error("KISS serial enabled, but no device is set.")
                    raise exception.MissingConfigOptionException(
                        "kiss_serial.device is not set.",
                    )
            elif transport == client.TRANSPORT_TCPKISS:
                if not CONF.kiss_tcp.host:
                    LOG.error("KISS TCP enabled, but no host is set.")
                    raise exception.MissingConfigOptionException(
                        "kiss_tcp.host is not set.",
                    )

            return True
        return False

    def is_alive(self):
        if self._client:
            return self._client.is_alive()
        else:
            return False

    def close(self):
        if self._client:
            self._client.stop()

    @staticmethod
    def transport():
        if CONF.kiss_serial.enabled:
            return client.TRANSPORT_SERIALKISS

        if CONF.kiss_tcp.enabled:
            return client.TRANSPORT_TCPKISS

    def decode_packet(self, *args, **kwargs):
        """We get a frame, which has to be decoded."""
        LOG.debug(f"kwargs {kwargs}")
        frame = kwargs["frame"]
        LOG.debug(f"Got an APRS Frame '{frame}'")
        # try and nuke the * from the fromcall sign.
        # frame.header._source._ch = False
        # payload = str(frame.payload.decode())
        # msg = f"{str(frame.header)}:{payload}"
        # msg = frame.tnc2
        # LOG.debug(f"Decoding {msg}")

        raw = aprslib.parse(str(frame))
        packet = core.factory(raw)
        if isinstance(packet, core.ThirdPartyPacket):
            return packet.subpacket
        else:
            return packet

    def setup_connection(self):
        self._client = kiss.KISS3Client()
        self.connected = True
        return self._client

    def consumer(self, callback, blocking=False, immortal=False, raw=False):
        self._client.consumer(callback)



import datetime
import email
from email.mime.text import MIMEText
import imaplib
import logging
import re
import smtplib
import threading
import time

import imapclient
from oslo_config import cfg

from aprsd import packets, plugin, threads, utils
from aprsd.threads import tx
from aprsd.utils import trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")
shortcuts_dict = None


class EmailInfo:
    """A singleton thread safe mechanism for the global check_email_delay.

    This has to be done because we have 2 separate threads that access
    the delay value.
    1) when EmailPlugin runs from a user message and
    2) when the background EmailThread runs to check email.

    Access the check email delay with
    EmailInfo().delay

    Set it with
    EmailInfo().delay = 100
    or
    EmailInfo().delay += 10

    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """This magic turns this into a singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.lock = threading.Lock()
            cls._instance._delay = 60
        return cls._instance

    @property
    def delay(self):
        with self.lock:
            return self._delay

    @delay.setter
    def delay(self, val):
        with self.lock:
            self._delay = val


@utils.singleton
class EmailStats:
    """Singleton object to store stats related to email."""
    _instance = None
    tx = 0
    rx = 0
    email_thread_last_time = None

    def stats(self, serializable=False):
        if CONF.email_plugin.enabled:
            last_check_time = self.email_thread_last_time
            if serializable and last_check_time:
                last_check_time = last_check_time.isoformat()
            stats = {
                "tx": self.tx,
                "rx": self.rx,
                "last_check_time": last_check_time,
            }
        else:
            stats = {}
        return stats

    def tx_inc(self):
        self.tx += 1

    def rx_inc(self):
        self.rx += 1

    def email_thread_update(self):
        self.email_thread_last_time = datetime.datetime.now()


class EmailPlugin(plugin.APRSDRegexCommandPluginBase):
    """Email Plugin."""

    command_regex = "^-.*"
    command_name = "email"
    short_description = "Send and Receive email"

    # message_number:time combos so we don't resend the same email in
    # five mins {int:int}
    email_sent_dict = {}
    enabled = False

    def setup(self):
        """Ensure that email is enabled and start the thread."""
        if CONF.email_plugin.enabled:
            self.enabled = True

            if not CONF.email_plugin.callsign:
                self.enabled = False
                LOG.error("email_plugin.callsign is not set.")
                return

            if not CONF.email_plugin.imap_login:
                LOG.error("email_plugin.imap_login not set. Disabling Plugin")
                self.enabled = False
                return

            if not CONF.email_plugin.smtp_login:
                LOG.error("email_plugin.smtp_login not set. Disabling Plugin")
                self.enabled = False
                return

            shortcuts = _build_shortcuts_dict()
            LOG.info(f"Email shortcuts {shortcuts}")
        else:
            LOG.info("Email services not enabled.")
            self.enabled = False

    def create_threads(self):
        if self.enabled:
            return APRSDEmailThread()

    @trace.trace
    def process(self, packet: packets.MessagePacket):
        LOG.info("Email COMMAND")
        if not self.enabled:
            # Email has not been enabled
            # so the plugin will just NOOP
            return packets.NULL_MESSAGE

        fromcall = packet.from_call
        message = packet.message_text
        ack = packet.get("msgNo", "0")

        reply = None
        if not CONF.email_plugin.enabled:
            LOG.debug("Email is not enabled in config file ignoring.")
            return "Email not enabled."

        searchstring = "^" + CONF.email_plugin.callsign + ".*"
        # only I can do email
        if re.search(searchstring, fromcall):
            # digits only, first one is number of emails to resend
            r = re.search("^-([0-9])[0-9]*$", message)
            if r is not None:
                LOG.debug("RESEND EMAIL")
                resend_email(r.group(1), fromcall)
                reply = packets.NULL_MESSAGE
            # -user@address.com body of email
            elif re.search(r"^-([A-Za-z0-9_\-\.@]+) (.*)", message):
                # (same search again)
                a = re.search(r"^-([A-Za-z0-9_\-\.@]+) (.*)", message)
                if a is not None:
                    to_addr = a.group(1)
                    content = a.group(2)

                    email_address = get_email_from_shortcut(to_addr)
                    if not email_address:
                        reply = "Bad email address"
                        return reply

                    # send recipient link to aprs.fi map
                    if content == "mapme":
                        content = (
                            "Click for my location: http://aprs.fi/{}" ""
                        ).format(
                            CONF.email_plugin.callsign,
                        )
                    too_soon = 0
                    now = time.time()
                    # see if we sent this msg number recently
                    if ack in self.email_sent_dict:
                        # BUG(hemna) - when we get a 2 different email command
                        # with the same ack #, we don't send it.
                        timedelta = now - self.email_sent_dict[ack]
                        if timedelta < 300:  # five minutes
                            too_soon = 1
                    if not too_soon or ack == 0:
                        LOG.info(f"Send email '{content}'")
                        send_result = send_email(to_addr, content)
                        reply = packets.NULL_MESSAGE
                        if send_result != 0:
                            reply = f"-{to_addr} failed"
                        else:
                            # clear email sent dictionary if somehow goes
                            # over 100
                            if len(self.email_sent_dict) > 98:
                                LOG.debug(
                                    "DEBUG: email_sent_dict is big ("
                                    + str(len(self.email_sent_dict))
                                    + ") clearing out.",
                                )
                                self.email_sent_dict.clear()
                            self.email_sent_dict[ack] = now
                    else:
                        reply = packets.NULL_MESSAGE
                        LOG.info(
                            "Email for message number "
                            + ack
                            + " recently sent, not sending again.",
                        )
            else:
                reply = "Bad email address"

        return reply


def _imap_connect():
    imap_port = CONF.email_plugin.imap_port
    use_ssl = CONF.email_plugin.imap_use_ssl

    try:
        server = imapclient.IMAPClient(
            CONF.email_plugin.imap_host,
            port=imap_port,
            use_uid=True,
            ssl=use_ssl,
            timeout=30,
        )
    except Exception:
        LOG.exception("Failed to connect IMAP server")
        return

    try:
        server.login(
            CONF.email_plugin.imap_login,
            CONF.email_plugin.imap_password,
        )
    except (imaplib.IMAP4.error, Exception) as e:
        msg = getattr(e, "message", repr(e))
        LOG.error(f"Failed to login {msg}")
        return

    server.select_folder("INBOX")

    server.fetch = trace.trace(server.fetch)
    server.search = trace.trace(server.search)
    server.remove_flags = trace.trace(server.remove_flags)
    server.add_flags = trace.trace(server.add_flags)
    return server


def _smtp_connect():
    host = CONF.email_plugin.smtp_host
    smtp_port = CONF.email_plugin.smtp_port
    use_ssl = CONF.email_plugin.smtp_use_ssl
    msg = "{}{}:{}".format("SSL " if use_ssl else "", host, smtp_port)
    LOG.debug(
        "Connect to SMTP host {} with user '{}'".format(
            msg,
            CONF.email_plugin.smtp_login,
        ),
    )

    try:
        if use_ssl:
            server = smtplib.SMTP_SSL(
                host=host,
                port=smtp_port,
                timeout=30,
            )
        else:
            server = smtplib.SMTP(
                host=host,
                port=smtp_port,
                timeout=30,
            )
    except Exception:
        LOG.error("Couldn't connect to SMTP Server")
        return

    LOG.debug(f"Connected to smtp host {msg}")

    debug = CONF.email_plugin.debug
    if debug:
        server.set_debuglevel(5)
        server.sendmail = trace.trace(server.sendmail)

    try:
        server.login(
            CONF.email_plugin.smtp_login,
            CONF.email_plugin.smtp_password,
        )
    except Exception:
        LOG.error("Couldn't connect to SMTP Server")
        return

    LOG.debug(f"Logged into SMTP server {msg}")
    return server


def _build_shortcuts_dict():
    global shortcuts_dict
    if not shortcuts_dict:
        if CONF.email_plugin.email_shortcuts:
            shortcuts_dict = {}
            tmp = CONF.email_plugin.email_shortcuts
            for combo in tmp:
                entry = combo.split("=")
                shortcuts_dict[entry[0]] = entry[1]
        else:
            shortcuts_dict = {}

    return shortcuts_dict


def get_email_from_shortcut(addr):
    if CONF.email_plugin.email_shortcuts:
        shortcuts = _build_shortcuts_dict()
        LOG.info(f"Shortcut lookup {addr} returns {shortcuts.get(addr, addr)}")
        return shortcuts.get(addr, addr)
    else:
        return addr


def validate_email_config(disable_validation=False):
    """function to simply ensure we can connect to email services.

    This helps with failing early during startup.
    """
    LOG.info("Checking IMAP configuration")
    imap_server = _imap_connect()
    LOG.info("Checking SMTP configuration")
    smtp_server = _smtp_connect()

    if imap_server and smtp_server:
        return True
    else:
        return False


@trace.trace
def parse_email(msgid, data, server):
    envelope = data[b"ENVELOPE"]
    # email address match
    # use raw string to avoid invalid escape secquence errors r"string here"
    f = re.search(r"([\.\w_-]+@[\.\w_-]+)", str(envelope.from_[0]))
    if f is not None:
        from_addr = f.group(1)
    else:
        from_addr = "noaddr"
    LOG.debug(f"Got a message from '{from_addr}'")
    try:
        m = server.fetch([msgid], ["RFC822"])
    except Exception:
        LOG.exception("Couldn't fetch email from server in parse_email")
        return

    msg = email.message_from_string(m[msgid][b"RFC822"].decode(errors="ignore"))
    if msg.is_multipart():
        text = ""
        html = None
        # default in case body somehow isn't set below - happened once
        body = b"* unreadable msg received"
        # this uses the last text or html part in the email,
        # phone companies often put content in an attachment
        for part in msg.get_payload():
            if part.get_content_charset() is None:
                # or BREAK when we hit a text or html?
                # We cannot know the character set,
                # so return decoded "something"
                LOG.debug("Email got unknown content type")
                text = part.get_payload(decode=True)
                continue

            charset = part.get_content_charset()

            if part.get_content_type() == "text/plain":
                LOG.debug("Email got text/plain")
                text = str(
                    part.get_payload(decode=True),
                    str(charset),
                    "ignore",
                ).encode("utf8", "replace")

            if part.get_content_type() == "text/html":
                LOG.debug("Email got text/html")
                html = str(
                    part.get_payload(decode=True),
                    str(charset),
                    "ignore",
                ).encode("utf8", "replace")

            if text is not None:
                # strip removes white space fore and aft of string
                body = text.strip()
            else:
                body = html.strip()
    else:  # message is not multipart
        # email.uscc.net sends no charset, blows up unicode function below
        LOG.debug("Email is not multipart")
        if msg.get_content_charset() is None:
            text = str(msg.get_payload(decode=True), "US-ASCII", "ignore").encode(
                "utf8",
                "replace",
            )
        else:
            text = str(
                msg.get_payload(decode=True),
                msg.get_content_charset(),
                "ignore",
            ).encode("utf8", "replace")
        body = text.strip()

    # FIXED:  UnicodeDecodeError: 'ascii' codec can't decode byte 0xf0
    # in position 6: ordinal not in range(128)
    # decode with errors='ignore'.   be sure to encode it before we return
    # it below, also with errors='ignore'
    try:
        body = body.decode(errors="ignore")
    except Exception:
        LOG.exception("Unicode decode failure")
        LOG.error(f"Unidoce decode failed: {str(body)}")
        body = "Unreadable unicode msg"
    # strip all html tags
    body = re.sub("<[^<]+?>", "", body)
    # strip CR/LF, make it one line, .rstrip fails at this
    body = body.replace("\n", " ").replace("\r", " ")
    # ascii might be out of range, so encode it, removing any error characters
    body = body.encode(errors="ignore")
    return body, from_addr


# end parse_email


@trace.trace
def send_email(to_addr, content):
    shortcuts = _build_shortcuts_dict()
    email_address = get_email_from_shortcut(to_addr)
    LOG.info("Sending Email_________________")

    if to_addr in shortcuts:
        LOG.info(f"To          : {to_addr}")
        to_addr = email_address
        LOG.info(f" ({to_addr})")
    subject = CONF.email_plugin.callsign
    # content = content + "\n\n(NOTE: reply with one line)"
    LOG.info(f"Subject     : {subject}")
    LOG.info(f"Body        : {content}")

    # check email more often since there's activity right now
    EmailInfo().delay = 60

    msg = MIMEText(content)
    msg["Subject"] = subject
    msg["From"] = CONF.email_plugin.smtp_login
    msg["To"] = to_addr
    server = _smtp_connect()
    if server:
        try:
            server.sendmail(
                CONF.email_plugin.smtp_login,
                [to_addr],
                msg.as_string(),
            )
            EmailStats().tx_inc()
        except Exception:
            LOG.exception("Sendmail Error!!!!")
            server.quit()
            return -1
        server.quit()
        return 0


@trace.trace
def resend_email(count, fromcall):
    date = datetime.datetime.now()
    month = date.strftime("%B")[:3]  # Nov, Mar, Apr
    day = date.day
    year = date.year
    today = f"{day}-{month}-{year}"

    shortcuts = _build_shortcuts_dict()
    # swap key/value
    shortcuts_inverted = {v: k for k, v in shortcuts.items()}

    try:
        server = _imap_connect()
    except Exception:
        LOG.exception("Failed to Connect to IMAP. Cannot resend email ")
        return

    try:
        messages = server.search(["SINCE", today])
    except Exception:
        LOG.exception("Couldn't search for emails in resend_email ")
        return

    # LOG.debug("%d messages received today" % len(messages))

    msgexists = False

    messages.sort(reverse=True)
    del messages[int(count) :]  # only the latest "count" messages
    for message in messages:
        try:
            parts = server.fetch(message, ["ENVELOPE"]).items()
        except Exception:
            LOG.exception("Couldn't fetch email parts in resend_email")
            continue

        for msgid, data in list(parts):
            # one at a time, otherwise order is random
            (body, from_addr) = parse_email(msgid, data, server)
            # unset seen flag, will stay bold in email client
            try:
                server.remove_flags(msgid, [imapclient.SEEN])
            except Exception:
                LOG.exception("Failed to remove SEEN flag in resend_email")

            if from_addr in shortcuts_inverted:
                # reverse lookup of a shortcut
                from_addr = shortcuts_inverted[from_addr]
            # asterisk indicates a resend
            reply = "-" + from_addr + " * " + body.decode(errors="ignore")
            tx.send(
                packets.MessagePacket(
                    from_call=CONF.callsign,
                    to_call=fromcall,
                    message_text=reply,
                ),
            )
            msgexists = True

    if msgexists is not True:
        stm = time.localtime()
        h = stm.tm_hour
        m = stm.tm_min
        s = stm.tm_sec
        # append time as a kind of serial number to prevent FT1XDR from
        # thinking this is a duplicate message.
        # The FT1XDR pretty much ignores the aprs message number in this
        # regard.  The FTM400 gets it right.
        reply = "No new msg {}:{}:{}".format(
            str(h).zfill(2),
            str(m).zfill(2),
            str(s).zfill(2),
        )
        tx.send(
            packets.MessagePacket(
                from_call=CONF.callsign,
                to_call=fromcall,
                message_text=reply,
            ),
        )

    # check email more often since we're resending one now
    EmailInfo().delay = 60

    server.logout()
    # end resend_email()


class APRSDEmailThread(threads.APRSDThread):
    def __init__(self):
        super().__init__("EmailThread")
        self.past = datetime.datetime.now()

    def loop(self):
        time.sleep(5)
        EmailStats().email_thread_update()
        # always sleep for 5 seconds and see if we need to check email
        # This allows CTRL-C to stop the execution of this loop sooner
        # than check_email_delay time
        now = datetime.datetime.now()
        if now - self.past > datetime.timedelta(seconds=EmailInfo().delay):
            # It's time to check email

            # slowly increase delay every iteration, max out at 300 seconds
            # any send/receive/resend activity will reset this to 60 seconds
            if EmailInfo().delay < 300:
                EmailInfo().delay += 10
            LOG.debug(
                f"check_email_delay is {EmailInfo().delay} seconds ",
            )

            shortcuts = _build_shortcuts_dict()
            # swap key/value
            shortcuts_inverted = {v: k for k, v in shortcuts.items()}

            date = datetime.datetime.now()
            month = date.strftime("%B")[:3]  # Nov, Mar, Apr
            day = date.day
            year = date.year
            today = f"{day}-{month}-{year}"

            try:
                server = _imap_connect()
            except Exception:
                LOG.exception("IMAP Failed to connect")
                return True

            try:
                messages = server.search(["SINCE", today])
            except Exception:
                LOG.exception("IMAP failed to search for messages since today.")
                return True
            LOG.debug(f"{len(messages)} messages received today")

            try:
                _msgs = server.fetch(messages, ["ENVELOPE"])
            except Exception:
                LOG.exception("IMAP failed to fetch/flag messages: ")
                return True

            for msgid, data in _msgs.items():
                envelope = data[b"ENVELOPE"]
                LOG.debug(
                    'ID:%d  "%s" (%s)'
                    % (msgid, envelope.subject.decode(), envelope.date),
                )
                f = re.search(
                    r"'([[A-a][0-9]_-]+@[[A-a][0-9]_-\.]+)",
                    str(envelope.from_[0]),
                )
                if f is not None:
                    from_addr = f.group(1)
                else:
                    from_addr = "noaddr"

                # LOG.debug("Message flags/tags:  " +
                # str(server.get_flags(msgid)[msgid]))
                # if "APRS" not in server.get_flags(msgid)[msgid]:
                # in python3, imap tags are unicode.  in py2 they're strings.
                # so .decode them to handle both
                try:
                    taglist = [
                        x.decode(errors="ignore")
                        for x in server.get_flags(msgid)[msgid]
                    ]
                except Exception:
                    LOG.error("Failed to get flags.")
                    break

                if "APRS" not in taglist:
                    # if msg not flagged as sent via aprs
                    try:
                        server.fetch([msgid], ["RFC822"])
                    except Exception:
                        LOG.exception("Failed single server fetch for RFC822")
                        break

                    (body, from_addr) = parse_email(msgid, data, server)
                    # unset seen flag, will stay bold in email client
                    try:
                        server.remove_flags(msgid, [imapclient.SEEN])
                    except Exception:
                        LOG.exception("Failed to remove flags SEEN")
                        # Not much we can do here, so lets try and
                        # send the aprs message anyway

                    if from_addr in shortcuts_inverted:
                        # reverse lookup of a shortcut
                        from_addr = shortcuts_inverted[from_addr]

                    reply = "-" + from_addr + " " + body.decode(errors="ignore")
                    # Send the message to the registered user in the
                    # config ham.callsign
                    tx.send(
                        packets.MessagePacket(
                            from_call=CONF.callsign,
                            to_call=CONF.email_plugin.callsign,
                            message_text=reply,
                        ),
                    )
                    # flag message as sent via aprs
                    try:
                        server.add_flags(msgid, ["APRS"])
                        # unset seen flag, will stay bold in email client
                    except Exception:
                        LOG.exception("Couldn't add APRS flag to email")

                    try:
                        server.remove_flags(msgid, [imapclient.SEEN])
                    except Exception:
                        LOG.exception("Couldn't remove seen flag from email")

                    # check email more often since we just received an email
                    EmailInfo().delay = 60

            # reset clock
            LOG.debug("Done looping over Server.fetch, log out.")
            self.past = datetime.datetime.now()
            try:
                server.logout()
            except Exception:
                LOG.exception("IMAP failed to logout: ")
                return True
        else:
            # We haven't hit the email delay yet.
            # LOG.debug("Delta({}) < {}".format(now - past, check_email_delay))
            return True

        return True
